#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Test categorize_dae_variables_and_constraints function.
"""

import pytest
import pyomo.environ as pyo
import pyomo.dae as dae
from pyomo.common.collections import ComponentSet
from pyomo.dae.flatten import flatten_dae_components

from idaes.apps.caprese.categorize import (
    categorize_dae_variables_and_constraints,
)
from idaes.apps.caprese.common.config import VariableCategory as VC
from idaes.apps.caprese.common.config import ConstraintCategory as CC
from idaes.apps.caprese.tests.test_simple_model import make_model

__author__ = "Robert Parker"


@pytest.mark.unit
def test_categorize_deriv():
    """The simplest test. Identify a differential and a derivative var."""
    m = pyo.ConcreteModel()
    m.time = dae.ContinuousSet(initialize=[0, 1])
    m.v = pyo.Var(m.time, initialize=0)
    m.dv = dae.DerivativeVar(m.v, wrt=m.time)
    m.diff_eqn = pyo.Constraint(
        m.time, rule={t: m.dv[t] == -m.v[t] ** 2 for t in m.time}
    )
    with pytest.raises(TypeError):
        # If we find a derivative var, we will try to access the disc eq.
        var_partition, con_partition = categorize_dae_variables_and_constraints(
            m,
            [m.v, m.dv],
            [m.diff_eqn],
            m.time,
        )
    disc = pyo.TransformationFactory("dae.finite_difference")
    disc.apply_to(m, wrt=m.time, nfe=1, scheme="BACKWARD")
    var_partition, con_partition = categorize_dae_variables_and_constraints(
        m,
        [m.v, m.dv],
        [m.diff_eqn, m.dv_disc_eq],
        m.time,
    )
    assert len(var_partition[VC.DIFFERENTIAL]) == 1
    assert var_partition[VC.DIFFERENTIAL][0] is m.v
    assert len(var_partition[VC.DERIVATIVE]) == 1
    assert var_partition[VC.DERIVATIVE][0] is m.dv
    assert len(con_partition[CC.DIFFERENTIAL]) == 1
    assert con_partition[CC.DIFFERENTIAL][0] is m.diff_eqn
    assert len(con_partition[CC.DISCRETIZATION]) == 1
    assert con_partition[CC.DISCRETIZATION][0] is m.dv_disc_eq

    for categ in VC:
        if (
            categ is not VC.DIFFERENTIAL
            and categ is not VC.DERIVATIVE
            and categ in var_partition
        ):
            assert len(var_partition[categ]) == 0
    for categ in CC:
        if (
            categ is not CC.DIFFERENTIAL
            and categ is not CC.DISCRETIZATION
            and categ in con_partition
        ):
            assert len(con_partition[categ]) == 0


@pytest.mark.unit
def test_categorize_deriv_fixed():
    """If one of the derivative or diff var are fixed, the other
    should be categorized as algebraic.
    """
    m = pyo.ConcreteModel()
    m.time = dae.ContinuousSet(initialize=[0, 1])
    m.v = pyo.Var(m.time, initialize=0)
    m.dv = dae.DerivativeVar(m.v, wrt=m.time)
    m.diff_eqn = pyo.Constraint(
        m.time, rule={t: m.dv[t] == -m.v[t] ** 2 for t in m.time}
    )
    disc = pyo.TransformationFactory("dae.finite_difference")
    disc.apply_to(m, wrt=m.time, nfe=1, scheme="BACKWARD")

    #
    # Fix differential variable, e.g. it is an input
    #
    m.v.fix()
    m.diff_eqn.deactivate()

    var_partition, con_partition = categorize_dae_variables_and_constraints(
        m,
        [m.v, m.dv],
        [m.diff_eqn, m.dv_disc_eq],
        m.time,
    )
    # Expected categories have expected variables
    assert len(var_partition[VC.ALGEBRAIC]) == 1
    assert var_partition[VC.ALGEBRAIC][0] is m.dv
    assert len(con_partition[CC.ALGEBRAIC]) == 1
    assert con_partition[CC.ALGEBRAIC][0] is m.dv_disc_eq

    # Unexpected categories are empty
    for categ in VC:
        if (
            categ is not VC.ALGEBRAIC
            and categ is not VC.UNUSED
            and categ in var_partition
        ):
            assert len(var_partition[categ]) == 0
    for categ in CC:
        if (
            categ is not CC.ALGEBRAIC
            and categ is not CC.UNUSED
            and categ in con_partition
        ):
            assert len(con_partition[categ]) == 0

    #
    # We can accomplish something similar by making m.v an input
    #
    m.v.unfix()
    var_partition, con_partition = categorize_dae_variables_and_constraints(
        m,
        [m.v, m.dv],
        [m.diff_eqn, m.dv_disc_eq],
        m.time,
        input_vars=[m.v],
    )
    # Expected categories have expected variables
    assert len(var_partition[VC.ALGEBRAIC]) == 1
    assert var_partition[VC.ALGEBRAIC][0] is m.dv
    assert len(var_partition[VC.INPUT]) == 1
    assert var_partition[VC.INPUT][0] is m.v
    assert len(con_partition[CC.ALGEBRAIC]) == 1
    assert con_partition[CC.ALGEBRAIC][0] is m.dv_disc_eq

    #
    # Fix derivative var, e.g. pseudo-steady state
    #
    for var in m.dv.values():
        var.fix(0.0)
    m.diff_eqn.activate()
    m.dv_disc_eq.deactivate()
    var_partition, con_partition = categorize_dae_variables_and_constraints(
        m,
        [m.v, m.dv],
        [m.diff_eqn, m.dv_disc_eq],
        m.time,
    )
    # Expected categories have expected variables
    assert len(var_partition[VC.ALGEBRAIC]) == 1
    assert var_partition[VC.ALGEBRAIC][0] is m.v
    assert len(con_partition[CC.ALGEBRAIC]) == 1
    assert con_partition[CC.ALGEBRAIC][0] is m.diff_eqn


@pytest.mark.unit
def test_categorize_simple_model():
    """Categorize variables and equations in the "simple model" used
    for the base class unit tests.
    """
    m = make_model()
    m.conc_in.unfix()
    m.flow_in.unfix()
    scalar_vars, dae_vars = flatten_dae_components(m, m.time, pyo.Var)
    scalar_cons, dae_cons = flatten_dae_components(m, m.time, pyo.Constraint)
    var_partition, con_partition = categorize_dae_variables_and_constraints(
        m,
        dae_vars,
        dae_cons,
        m.time,
        input_vars=[m.flow_in],
        disturbance_vars=[
            pyo.Reference(m.conc_in[:, "A"]),
            pyo.Reference(m.conc_in[:, "B"]),
        ],
    )
    t1 = m.time.at(2)
    # Expected variables:
    expected_vars = {
        VC.DIFFERENTIAL: ComponentSet([m.conc[t1, "A"], m.conc[t1, "B"]]),
        VC.DERIVATIVE: ComponentSet([m.dcdt[t1, "A"], m.dcdt[t1, "B"]]),
        VC.ALGEBRAIC: ComponentSet(
            [
                m.rate[t1, "A"],
                m.rate[t1, "B"],
                m.flow_out[t1],
            ]
        ),
        VC.INPUT: ComponentSet([m.flow_in[t1]]),
        VC.DISTURBANCE: ComponentSet(
            [
                m.conc_in[t1, "A"],
                m.conc_in[t1, "B"],
            ]
        ),
    }

    # Expected constraints:
    expected_cons = {
        CC.DIFFERENTIAL: ComponentSet(
            [
                m.material_balance[t1, "A"],
                m.material_balance[t1, "B"],
            ]
        ),
        CC.DISCRETIZATION: ComponentSet(
            [
                m.dcdt_disc_eq[t1, "A"],
                m.dcdt_disc_eq[t1, "B"],
            ]
        ),
        CC.ALGEBRAIC: ComponentSet(
            [
                m.rate_eqn[t1, "A"],
                m.rate_eqn[t1, "B"],
                m.flow_eqn[t1],
            ]
        ),
    }

    # Expected categories have expected variables and constraints
    for categ in expected_vars:
        assert len(expected_vars[categ]) == len(var_partition[categ])
        for var in var_partition[categ]:
            assert var[t1] in expected_vars[categ]
    for categ in var_partition:
        if categ not in expected_vars:
            assert len(var_partition[categ]) == 0

    for categ in expected_cons:
        assert len(expected_cons[categ]) == len(con_partition[categ])
        for con in con_partition[categ]:
            assert con[t1] in expected_cons[categ]
    for categ in con_partition:
        if categ not in expected_cons:
            assert len(con_partition[categ]) == 0


@pytest.mark.unit
def test_categorize_simple_model_with_constraints():
    """Categorize variables and equations in the "simple model" used
    for the base class unit tests, now with an "input constraint"
    and an active inequality.

    Note that the active inequality is in terms of a differential var,
    so that var must be matched with the inequality and therefore
    must be "considered differential."
    """
    m = make_model()
    m.conc_in.unfix()
    m.flow_in.unfix()
    m.performance = pyo.Constraint(
        m.time, rule={t: m.conc[t, "A"] >= 0.5 for t in m.time}
    )
    m.pwc_input = pyo.Constraint(m.time, rule={t: m.flow_in[t] == 1.0 for t in m.time})
    scalar_vars, dae_vars = flatten_dae_components(m, m.time, pyo.Var)
    scalar_cons, dae_cons = flatten_dae_components(m, m.time, pyo.Constraint)
    var_partition, con_partition = categorize_dae_variables_and_constraints(
        m,
        dae_vars,
        dae_cons,
        m.time,
        input_vars=[m.flow_in],
        disturbance_vars=[pyo.Reference(m.conc_in[:, "B"])],
        input_cons=[m.pwc_input],
        active_inequalities=[m.performance],
    )
    t1 = m.time.at(2)
    # Expected variables:
    expected_vars = {
        VC.DIFFERENTIAL: ComponentSet([m.conc[t1, "B"]]),
        VC.DERIVATIVE: ComponentSet([m.dcdt[t1, "B"]]),
        VC.ALGEBRAIC: ComponentSet(
            [
                m.rate[t1, "A"],
                m.rate[t1, "B"],
                m.flow_out[t1],
                m.conc[t1, "A"],
                m.dcdt[t1, "A"],
                m.conc_in[t1, "A"],
            ]
        ),
        VC.INPUT: ComponentSet([m.flow_in[t1]]),
        VC.DISTURBANCE: ComponentSet(
            [
                m.conc_in[t1, "B"],
            ]
        ),
    }

    # Expected constraints:
    expected_cons = {
        CC.DIFFERENTIAL: ComponentSet(
            [
                m.material_balance[t1, "B"],
            ]
        ),
        CC.DISCRETIZATION: ComponentSet(
            [
                m.dcdt_disc_eq[t1, "B"],
            ]
        ),
        CC.ALGEBRAIC: ComponentSet(
            [
                m.rate_eqn[t1, "A"],
                m.rate_eqn[t1, "B"],
                m.flow_eqn[t1],
                m.material_balance[t1, "A"],
                m.dcdt_disc_eq[t1, "A"],
                m.performance[t1],
            ]
        ),
        CC.INPUT: ComponentSet(
            [
                m.pwc_input[t1],
            ]
        ),
    }

    # Expected categories have expected variables and constraints
    for categ in expected_vars:
        assert len(expected_vars[categ]) == len(var_partition[categ])
        for var in var_partition[categ]:
            assert var[t1] in expected_vars[categ]
    for categ in var_partition:
        if categ not in expected_vars:
            assert len(var_partition[categ]) == 0

    for categ in expected_cons:
        assert len(expected_cons[categ]) == len(con_partition[categ])
        for con in con_partition[categ]:
            assert con[t1] in expected_cons[categ]
    for categ in con_partition:
        if categ not in expected_cons:
            assert len(con_partition[categ]) == 0


@pytest.mark.unit
def test_space_indexed_model():
    """The case where a differential variable is "specified"
    by an input (boundary condition). This is the case that
    motivated the development of this categorization method.
    """
    m = pyo.ConcreteModel()
    m.time = dae.ContinuousSet(initialize=[0, 1])
    m.space = dae.ContinuousSet(initialize=[0, 1])
    m.v = pyo.Var(m.time, m.space, initialize=0)

    m.dvdt = dae.DerivativeVar(m.v, wrt=m.time)
    m.dvdx = dae.DerivativeVar(m.v, wrt=m.space)

    def _balance_rule(m, t, x):
        return m.dvdt[t, x] + m.dvdx[t, x] == 0

    m.balance = pyo.Constraint(m.time, m.space, rule=_balance_rule)

    m.u = pyo.Var(m.time, m.space, initialize=0)

    def _u_eqn_rule(m, t, x):
        return m.v[t, x] == m.u[t, x]

    m.u_eqn = pyo.Constraint(m.time, m.space, rule=_u_eqn_rule)

    disc = pyo.TransformationFactory("dae.finite_difference")
    disc.apply_to(m, wrt=m.time, nfe=2, scheme="BACKWARD")
    disc.apply_to(m, wrt=m.space, nfe=2, scheme="BACKWARD")

    scalar_vars, dae_vars = flatten_dae_components(m, m.time, pyo.Var)
    scalar_cons, dae_cons = flatten_dae_components(m, m.time, pyo.Constraint)
    var_partition, con_partition = categorize_dae_variables_and_constraints(
        m,
        dae_vars,
        dae_cons,
        m.time,
        input_vars=[pyo.Reference(m.u[:, 0])],
    )

    t1 = m.time.at(2)
    # Expected variables:
    expected_vars = {
        VC.DIFFERENTIAL: ComponentSet([m.v[t1, x] for x in m.space if x != 0]),
        VC.DERIVATIVE: ComponentSet([m.dvdt[t1, x] for x in m.space if x != 0]),
        VC.ALGEBRAIC: ComponentSet(
            [
                m.dvdt[t1, 0],
                m.v[t1, 0],
                *[m.dvdx[t1, x] for x in m.space],
                *[m.u[t1, x] for x in m.space if x != 0],
            ]
        ),
        VC.INPUT: ComponentSet([m.u[t1, 0]]),
    }

    # Expected constraints:
    expected_cons = {
        CC.DIFFERENTIAL: ComponentSet([m.balance[t1, x] for x in m.space if x != 0]),
        CC.DISCRETIZATION: ComponentSet(
            [m.dvdt_disc_eq[t1, x] for x in m.space if x != 0]
        ),
        CC.ALGEBRAIC: ComponentSet(
            [
                m.balance[t1, 0],
                m.dvdt_disc_eq[t1, 0],
                *[m.dvdx_disc_eq[t1, x] for x in m.space if x != 0],
                *[m.u_eqn[t1, x] for x in m.space],
            ]
        ),
    }

    # Expected categories have expected variables and constraints
    for categ in expected_vars:
        assert len(expected_vars[categ]) == len(var_partition[categ])
        for var in var_partition[categ]:
            assert var[t1] in expected_vars[categ]
    for categ in var_partition:
        if categ not in expected_vars:
            assert len(var_partition[categ]) == 0

    for categ in expected_cons:
        assert len(expected_cons[categ]) == len(con_partition[categ])
        for con in con_partition[categ]:
            assert con[t1] in expected_cons[categ]
    for categ in con_partition:
        if categ not in expected_cons:
            assert len(con_partition[categ]) == 0



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

import importlib.abc
import importlib.machinery
import os
import subprocess
import sys
from collections.abc import Iterable
from pathlib import Path
from typing import Union

import pytest

####
# Uncomment this to collect list of all test files
# into 'test_files.txt'
#
# def pytest_collection_modifyitems(config, items):
#     output = open("test_files.txt", "w")
#     fspaths = set()
#     for item in items:
#         fspaths.add(item.fspath)
#     for p in fspaths:
#         output.write(str(p))
#         output.write("\n")
#     output.close()
####

####


def pytest_addoption(parser):
    parser.addoption(
        "--performance",
        action="store_true",
        dest="performance",
        default=False,
        help="enable performance decorated tests",
    )


MARKERS = {
    "build": "test of model build methods",
    "cubic_root": "test requires the compiled cubic root finder",
    "iapws": "test requires the compiled IAPWS95 property package",
    "initialization": "test of initialization methods. These generally require a solver as well",
    "solver": "test requires a solver",
    "ui": "tests of an aspect of the ui",
    "unit": "quick tests that do not require a solver, must run in <2s",
    "component": "quick tests that may require a solver",
    "integration": "long duration tests",
    "performance": "tests for the IDAES performance testing suite",
}


def pytest_configure(config: pytest.Config):
    for name, description in MARKERS.items():
        config.addinivalue_line("markers", f"{name}: {description}")

    if not config.option.performance:
        if len(config.option.markexpr) > 0:
            setattr(
                config.option,
                "markexpr",
                f"{config.option.markexpr} and not performance",
            )
        else:
            setattr(config.option, "markexpr", "not performance")
    else:
        setattr(config.option, "markexpr", "performance")


REQUIRED_MARKERS = {"unit", "component", "integration", "performance"}
ALL_PLATFORMS = {"darwin", "linux", "win32"}


@pytest.hookimpl
def pytest_runtest_setup(item):
    _validate_required_markers(
        item, required_markers=REQUIRED_MARKERS, expected_count=1
    )
    _skip_for_unsupported_platforms(
        item,
        all_platforms=ALL_PLATFORMS,
        negate_tag=(lambda tag: f"no{tag}"),
    )


def _skip_for_unsupported_platforms(item, all_platforms=None, negate_tag=None):
    """
    Only run the tests for your particular platform(s), if it is marked with those platform(s)
    e.g. to mark tests for linux:

    import pytest
    @pytest.mark.linux
    def test_something():
       print("this only runs on linux")

    In addition, you can use "no<platform>" to exclude

    @pytest.mark.nowin32
    def test_something():
       print("this will not run on windows")

    The names of the platforms should match what is returned by `sys.platform`, in particular:
       Linux = 'linux'
       Windows = 'win32'
       macOS = 'darwin'
    """

    all_platforms = set(all_platforms or [])
    if negate_tag is None:

        def negate_tag(tag):
            return f"no{tag}"

    all_negated_platforms = {negate_tag(tag) for tag in all_platforms}
    item_markers = {marker.name for marker in item.iter_markers()}
    supported_platforms = all_platforms & item_markers
    excluded_platforms = all_negated_platforms & item_markers
    plat = sys.platform
    if (excluded_platforms and negate_tag(plat) in excluded_platforms) or (
        supported_platforms and plat not in supported_platforms
    ):
        pytest.skip(f"cannot run on platform {plat}")


def _validate_required_markers(item, required_markers=None, expected_count=1):
    required_markers = set(required_markers or [])
    item_markers = {marker.name for marker in item.iter_markers()}
    required_markers_on_item = item_markers & required_markers
    required_count = len(required_markers_on_item)
    reason_to_fail = None
    if required_count < expected_count:
        reason_to_fail = "Too few required markers"
    if required_count > expected_count:
        reason_to_fail = "Too many required markers"
    if reason_to_fail:
        msg = (
            f'{reason_to_fail} for test function "{item.name}". '
            f"Expected: {expected_count} of {required_markers}; "
            f"found: {required_markers_on_item or required_count}"
        )
        pytest.fail(msg)


ModuleName = str


class ImportorskipLoader(importlib.abc.Loader):
    """
    A wrapper class around a concrete Loader instance. If a ModuleNotFoundError is raised
    during module execution and the module name matches one of the registered modules,
    it is replaced with a module-level call to :func:`pyest.skip()`.
    """

    def __init__(
        self, wrapped: importlib.abc.Loader, skip_if_not_found: Iterable[ModuleName]
    ):
        self._wrapped = wrapped
        self.skip_if_not_found = list(skip_if_not_found)

    def module_repr(self, module) -> str:
        return self._wrapped.module_repr(module)

    def create_module(self, spec):
        return self._wrapped.create_module(spec)

    def exec_module(self, module):
        try:
            return self._wrapped.exec_module(module)
        except ModuleNotFoundError as e:
            if e.name in self.skip_if_not_found:
                pytest.skip(allow_module_level=True)
            raise e


class ImportorskipFinder(importlib.abc.MetaPathFinder):
    """
    Custom Finder class to modify import behavior for registered modules.

    If inserted in sys.meta_path before the default finders, it will cause
    a custom Loader to be used for registered modules.
    """

    def __init__(self, registry: dict[ModuleName, list[ModuleName]]):
        self._registry = registry

    def find_spec(self, *args, **kwargs):
        spec = importlib.machinery.PathFinder.find_spec(*args, **kwargs)
        if spec is None:
            return
        registered_for_skipping = self._registry.get(spec.name, None)
        if registered_for_skipping:
            spec.loader = ImportorskipLoader(spec.loader, registered_for_skipping)
        return spec


class Importorskipper:
    """
    A pytest plugin that allows automatically skipping test modules if they attempt to import
    modules distributed in optional dependencies.

    The functionality is similar to pytest.importorskip(), but using a centralized registry
    instead of having to call importorskip() in each test module for each possibly missing module.
    """

    def __init__(self, registry: dict[ModuleName, list[ModuleName]]):
        self._registry = dict(registry)
        self._finder = ImportorskipFinder(self._registry)

    def pytest_configure(self):
        sys.meta_path.insert(0, self._finder)

    def pytest_sessionfinish(self):
        sys.meta_path.remove(self._finder)

    def pytest_report_collectionfinish(self) -> list[str]:
        preamble = [
            "The following modules are registered in the importorskipper plugin",
            " and will cause tests to be skipped if any of the registered modules is not found: ",
        ]
        lines = []
        for importing_mod, mods in self._registry.items():
            lines.append(f"- {importing_mod}:\t{mods}")
        if lines:
            lines = preamble + lines
        return lines


@pytest.fixture(scope="function")
def run_in_tmp_path(tmp_path: Path):
    prev_workdir = os.getcwd()
    try:
        os.chdir(tmp_path)
        yield
    finally:
        os.chdir(prev_workdir)


@pytest.fixture(scope="class")
def run_class_in_tmp_path(
    request: pytest.FixtureRequest, tmp_path_factory: pytest.TempPathFactory
):
    prev_workdir = os.getcwd()

    modname = request.module.__name__
    clsname = request.__class__.__name__
    dirname = f"{modname}-{clsname}"

    tmp_path = tmp_path_factory.mktemp(dirname, numbered=True)
    try:
        os.chdir(tmp_path)
        yield
    finally:
        os.chdir(prev_workdir)


@pytest.fixture(scope="module")
def run_module_in_tmp_path(
    request: pytest.FixtureRequest, tmp_path_factory: pytest.TempPathFactory
):
    prev_workdir = os.getcwd()

    modname = request.module.__name__

    tmp_path = tmp_path_factory.mktemp(modname, numbered=True)
    try:
        os.chdir(tmp_path)
        yield
    finally:
        os.chdir(prev_workdir)


def _get_repo_root_dir() -> Union[Path, None]:
    try:
        text = subprocess.check_output(
            ["git", "rev-parse", "--show-toplevel"],
            text=True,
        ).strip()
    except subprocess.CalledProcessError:
        # no Git directory found
        return None
    return Path(text).resolve()


class VerifyCleanup:
    """
    A pytest plugin to verify that files generated by tests are cleaned up properly.
    """

    def __init__(self, repo_root_dir: Path):
        self._repo_root_dir = Path(repo_root_dir).resolve()
        self._added_by_test = {}

    def _get_files(self) -> list[str]:
        args = [
            "git",
            "-C",
            self._repo_root_dir,
            "ls-files",
            "--others",
            "--exclude-standard",
        ]
        try:
            text = subprocess.check_output(
                args,
                text=True,
            ).strip()
        except subprocess.CalledProcessError as e:
            text = str(e)
        return text.splitlines()

    def pytest_report_collectionfinish(self) -> list[str]:
        return [
            f"Looking for untracked files being created by tests for repo: {self._repo_root_dir}"
        ]

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_protocol(self, item):
        before = set(self._get_files())
        yield
        after = set(self._get_files())
        added = after - before
        if added:
            key = item.nodeid
            self._added_by_test[key] = sorted(added)

    @pytest.hookimpl(trylast=True)
    def pytest_terminal_summary(self, terminalreporter):
        to_report = dict(self._added_by_test)
        if not to_report:
            return

        tr = terminalreporter
        tr.section("Files added (and not cleaned up) by tests")
        for item, files in to_report.items():
            tr.write_line(item)
            for file in files:
                tr.write_line(f"\t{file}")
        if to_report:
            tr.write_line(f"{len(to_report)} test(s) did not clean up after themselves")
            tr.write_line("The exit status of the test session will be set to failed")

    @pytest.hookimpl(trylast=True)
    def pytest_sessionfinish(self, session, exitstatus):
        if self._added_by_test:
            session.exitstatus = pytest.ExitCode.TESTS_FAILED


def pytest_addhooks(pluginmanager: pytest.PytestPluginManager):
    skipper_plugin = Importorskipper(
        {
            "idaes.core.surrogate.keras_surrogate": ["omlt"],
        }
    )

    pluginmanager.register(skipper_plugin, name="importorskipper")
    repo_root_dir = _get_repo_root_dir()
    if repo_root_dir is not None:
        pluginmanager.register(VerifyCleanup(repo_root_dir))


####



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for the PIDController

Author: Douglas Allan
"""
import pytest
from io import StringIO

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    Expression,
    Param,
    value,
    Var,
    units as pyunits,
)
from pyomo.util.check_units import assert_units_consistent, assert_units_equivalent
from pyomo.dae import DerivativeVar

from idaes.core import FlowsheetBlock, EnergyBalanceType, MomentumBalanceType
from idaes.models.control import (
    PIDController,
    ControllerType,
    ControllerMVBoundType,
    ControllerAntiwindupType,
)
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    fixed_variables_set,
    activated_constraints_set,
    number_unused_variables,
)
from idaes.core.util.testing import PhysicalParameterTestBlock, initialization_tester
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import ConfigurationError
from idaes.models_extra.power_generation.unit_models.soc_submodels.testing import (
    _build_test_utility,
)


def common_components(nt):
    return {
        Var: {
            "setpoint": nt,
            "process_var": nt,
            "manipulated_var": nt,
            "mv_ref": nt,
            "gain_p": nt,
        },
        Constraint: {"mv_eqn": nt},
        Expression: {"mv_unbounded": nt},
        Param: {},
    }


def add_integral_components(comp_dict, nt):
    comp_dict[Var] = {
        **comp_dict[Var],
        **{
            "gain_i": nt,
            "mv_integral_component": nt,
            "mv_integral_component_dot": nt,
        },
    }
    comp_dict[Constraint] = {
        **comp_dict[Constraint],
        **{
            "mv_integration_eqn": nt,
        },
    }


def add_derivative_components(comp_dict, nt):
    comp_dict[Var]["gain_d"] = nt
    comp_dict[Var]["derivative_term"] = nt


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_not_dynamic_exception():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False, time_units=pyunits.dimensionless)

    m.fs.process_var = Var(m.fs.time, units=pyunits.dimensionless)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.dimensionless)

    with pytest.raises(
        ConfigurationError, match="PIDControllers work only with dynamic flowsheets."
    ):
        m.fs.unit = PIDController(
            process_var=m.fs.process_var, manipulated_var=m.fs.manipulated_var
        )


@pytest.mark.unit
def test_pi_control():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=True, time_set=[0, 1], time_units=pyunits.s)

    nt = 2  # Two timepoints

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var, manipulated_var=m.fs.manipulated_var
    )

    # Check unit config arguments
    assert len(unit.config) == 9

    assert unit.config.dynamic
    assert unit.config.has_holdup

    assert unit.process_var.referent is m.fs.process_var
    assert unit.manipulated_var.referent is m.fs.manipulated_var

    comp_dict = common_components(nt)
    add_integral_components(comp_dict, nt)
    comp_dict[Expression]["error"] = nt
    comp_dict[Constraint]["initial_integral_error_eqn"] = 1

    _build_test_utility(
        block=m.fs.unit,
        comp_dict=comp_dict,
        references=["process_var", "manipulated_var"],
    )
    assert unit.mv_integration_eqn[0].active is False


@pytest.mark.unit
def test_p_control():
    m = ConcreteModel()
    nt = 3  # Three timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var,
        manipulated_var=m.fs.manipulated_var,
        controller_type=ControllerType.P,
    )

    # Check unit config arguments
    assert len(unit.config) == 9

    assert unit.config.dynamic
    assert unit.config.has_holdup

    assert unit.process_var.referent is m.fs.process_var
    assert unit.manipulated_var.referent is m.fs.manipulated_var

    comp_dict = common_components(nt)
    comp_dict[Expression]["error"] = nt

    _build_test_utility(
        block=m.fs.unit,
        comp_dict=comp_dict,
        references=["process_var", "manipulated_var"],
    )


@pytest.mark.unit
def test_pd_control():
    m = ConcreteModel()
    nt = 5  # Five timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var,
        manipulated_var=m.fs.manipulated_var,
        controller_type=ControllerType.PD,
    )

    # Check unit config arguments
    assert len(unit.config) == 9

    assert unit.config.dynamic
    assert unit.config.has_holdup

    assert unit.process_var.referent is m.fs.process_var
    assert unit.manipulated_var.referent is m.fs.manipulated_var

    comp_dict = common_components(nt)
    add_derivative_components(comp_dict, nt)
    comp_dict[Expression]["error"] = nt
    comp_dict[Var]["negative_pv"] = nt
    comp_dict[Constraint]["negative_pv_eqn"] = nt

    _build_test_utility(
        block=m.fs.unit,
        comp_dict=comp_dict,
        references=["process_var", "manipulated_var"],
    )


@pytest.mark.unit
def test_pid_control():
    m = ConcreteModel()
    nt = 5  # Five timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var,
        manipulated_var=m.fs.manipulated_var,
        controller_type=ControllerType.PID,
    )

    # Check unit config arguments
    assert len(unit.config) == 9

    assert unit.config.dynamic
    assert unit.config.has_holdup

    assert unit.process_var.referent is m.fs.process_var
    assert unit.manipulated_var.referent is m.fs.manipulated_var

    comp_dict = common_components(nt)
    add_integral_components(comp_dict, nt)
    add_derivative_components(comp_dict, nt)
    comp_dict[Expression]["error"] = nt
    comp_dict[Constraint]["initial_integral_error_eqn"] = 1
    comp_dict[Var]["negative_pv"] = nt
    comp_dict[Constraint]["negative_pv_eqn"] = nt

    _build_test_utility(
        block=m.fs.unit,
        comp_dict=comp_dict,
        references=["process_var", "manipulated_var"],
    )


@pytest.mark.component
def test_pid_units_consistent():
    m = ConcreteModel()
    nt = 5  # Five timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var,
        manipulated_var=m.fs.manipulated_var,
        controller_type=ControllerType.PID,
    )
    assert_units_consistent(m)
    assert_units_equivalent(m.fs.unit.setpoint, pyunits.m)
    assert_units_equivalent(m.fs.unit.mv_ref, pyunits.K)


@pytest.mark.unit
def test_derivative_on_error():
    m = ConcreteModel()
    nt = 5  # Five timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var,
        manipulated_var=m.fs.manipulated_var,
        controller_type=ControllerType.PID,
        derivative_on_error=True,
    )

    # Check unit config arguments
    assert len(unit.config) == 9

    assert unit.config.dynamic
    assert unit.config.has_holdup

    assert unit.process_var.referent is m.fs.process_var
    assert unit.manipulated_var.referent is m.fs.manipulated_var

    comp_dict = common_components(nt)
    add_integral_components(comp_dict, nt)
    add_derivative_components(comp_dict, nt)
    comp_dict[Var]["error"] = nt
    comp_dict[Constraint]["initial_integral_error_eqn"] = 1
    comp_dict[Constraint]["error_eqn"] = nt

    _build_test_utility(
        block=m.fs.unit,
        comp_dict=comp_dict,
        references=["process_var", "manipulated_var"],
    )


@pytest.mark.component
def test_derivative_on_error_units_consistent():
    m = ConcreteModel()
    nt = 5  # Five timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var,
        manipulated_var=m.fs.manipulated_var,
        controller_type=ControllerType.PID,
        derivative_on_error=True,
    )

    assert_units_consistent(m)
    assert_units_equivalent(m.fs.unit.setpoint, pyunits.m)
    assert_units_equivalent(m.fs.unit.mv_ref, pyunits.K)


@pytest.mark.unit
def test_logistic_mv_bound():
    m = ConcreteModel()
    nt = 5  # Five timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var,
        manipulated_var=m.fs.manipulated_var,
        controller_type=ControllerType.PID,
        mv_bound_type=ControllerMVBoundType.LOGISTIC,
    )

    # Check unit config arguments
    assert len(unit.config) == 9

    assert unit.config.dynamic
    assert unit.config.has_holdup

    assert unit.process_var.referent is m.fs.process_var
    assert unit.manipulated_var.referent is m.fs.manipulated_var

    comp_dict = common_components(nt)
    add_integral_components(comp_dict, nt)
    add_derivative_components(comp_dict, nt)
    comp_dict[Expression]["error"] = nt
    comp_dict[Constraint]["initial_integral_error_eqn"] = 1
    comp_dict[Var]["negative_pv"] = nt
    comp_dict[Constraint]["negative_pv_eqn"] = nt
    comp_dict[Param] = {"mv_lb": 1, "mv_ub": 1, "logistic_bound_k": 1}

    _build_test_utility(
        block=m.fs.unit,
        comp_dict=comp_dict,
        references=["process_var", "manipulated_var"],
    )


@pytest.mark.component
def test_logistic_mv_bound_units_consistent():
    m = ConcreteModel()
    nt = 5  # Five timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var,
        manipulated_var=m.fs.manipulated_var,
        controller_type=ControllerType.PID,
        mv_bound_type=ControllerMVBoundType.LOGISTIC,
    )
    assert_units_equivalent(m.fs.unit.setpoint, pyunits.m)
    assert_units_equivalent(m.fs.unit.mv_ref, pyunits.K)
    assert_units_consistent(m)


@pytest.mark.unit
def test_smooth_mv_bound():
    m = ConcreteModel()
    nt = 5  # Five timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var,
        manipulated_var=m.fs.manipulated_var,
        controller_type=ControllerType.PD,
        mv_bound_type=ControllerMVBoundType.SMOOTH_BOUND,
    )

    # Check unit config arguments
    assert len(unit.config) == 9

    assert unit.config.dynamic
    assert unit.config.has_holdup

    assert unit.process_var.referent is m.fs.process_var
    assert unit.manipulated_var.referent is m.fs.manipulated_var

    comp_dict = common_components(nt)
    add_derivative_components(comp_dict, nt)
    comp_dict[Expression]["error"] = nt
    comp_dict[Var]["negative_pv"] = nt
    comp_dict[Constraint]["negative_pv_eqn"] = nt
    comp_dict[Param] = {"mv_lb": 1, "mv_ub": 1, "smooth_eps": 1}

    _build_test_utility(
        block=m.fs.unit,
        comp_dict=comp_dict,
        references=["process_var", "manipulated_var"],
    )


@pytest.mark.component
def test_smooth_mv_bound_units_consistent():
    m = ConcreteModel()
    nt = 5  # Five timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var,
        manipulated_var=m.fs.manipulated_var,
        controller_type=ControllerType.PD,
        mv_bound_type=ControllerMVBoundType.SMOOTH_BOUND,
    )
    assert_units_equivalent(m.fs.unit.setpoint, pyunits.m)
    assert_units_equivalent(m.fs.unit.mv_ref, pyunits.K)
    assert_units_consistent(m)


@pytest.mark.unit
def test_conditional_integration():
    m = ConcreteModel()
    nt = 5  # Five timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var,
        manipulated_var=m.fs.manipulated_var,
        controller_type=ControllerType.PI,
        mv_bound_type=ControllerMVBoundType.SMOOTH_BOUND,
        antiwindup_type=ControllerAntiwindupType.CONDITIONAL_INTEGRATION,
    )

    # Check unit config arguments
    assert len(unit.config) == 9

    assert unit.config.dynamic
    assert unit.config.has_holdup

    assert unit.process_var.referent is m.fs.process_var
    assert unit.manipulated_var.referent is m.fs.manipulated_var

    comp_dict = common_components(nt)
    add_integral_components(comp_dict, nt)
    comp_dict[Expression]["error"] = nt
    comp_dict[Constraint]["initial_integral_error_eqn"] = 1
    comp_dict[Param] = {
        "mv_lb": 1,
        "mv_ub": 1,
        "smooth_eps": 1,
        "conditional_integration_k": 1,
    }

    _build_test_utility(
        block=m.fs.unit,
        comp_dict=comp_dict,
        references=["process_var", "manipulated_var"],
    )


@pytest.mark.component
def test_conditional_integration_units_consistent():
    m = ConcreteModel()
    nt = 5  # Five timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var,
        manipulated_var=m.fs.manipulated_var,
        controller_type=ControllerType.PI,
        mv_bound_type=ControllerMVBoundType.SMOOTH_BOUND,
        antiwindup_type=ControllerAntiwindupType.CONDITIONAL_INTEGRATION,
    )
    assert_units_equivalent(m.fs.unit.setpoint, pyunits.m)
    assert_units_equivalent(m.fs.unit.mv_ref, pyunits.K)
    assert_units_consistent(m)


@pytest.mark.unit
def test_back_calculation():
    m = ConcreteModel()
    nt = 11  # Eleven timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var,
        manipulated_var=m.fs.manipulated_var,
        controller_type=ControllerType.PID,
        mv_bound_type=ControllerMVBoundType.LOGISTIC,
        antiwindup_type=ControllerAntiwindupType.BACK_CALCULATION,
    )

    # Check unit config arguments
    assert len(unit.config) == 9

    assert unit.config.dynamic
    assert unit.config.has_holdup

    assert unit.process_var.referent is m.fs.process_var
    assert unit.manipulated_var.referent is m.fs.manipulated_var

    comp_dict = common_components(nt)
    add_integral_components(comp_dict, nt)
    add_derivative_components(comp_dict, nt)
    comp_dict[Expression]["error"] = nt
    comp_dict[Constraint]["initial_integral_error_eqn"] = 1
    comp_dict[Var]["negative_pv"] = nt
    comp_dict[Constraint]["negative_pv_eqn"] = nt
    comp_dict[Var]["gain_b"] = nt
    comp_dict[Param] = {"mv_lb": 1, "mv_ub": 1, "logistic_bound_k": 1}

    _build_test_utility(
        block=m.fs.unit,
        comp_dict=comp_dict,
        references=["process_var", "manipulated_var"],
    )


@pytest.mark.component
def test_back_calculation_units_consistent():
    m = ConcreteModel()
    nt = 11  # Eleven timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    unit = m.fs.unit = PIDController(
        process_var=m.fs.process_var,
        manipulated_var=m.fs.manipulated_var,
        controller_type=ControllerType.PID,
        mv_bound_type=ControllerMVBoundType.LOGISTIC,
        antiwindup_type=ControllerAntiwindupType.BACK_CALCULATION,
    )
    assert_units_equivalent(m.fs.unit.setpoint, pyunits.m)
    assert_units_equivalent(m.fs.unit.mv_ref, pyunits.K)
    assert_units_consistent(m)


@pytest.mark.unit
def test_antiwindup_no_bounds():
    m = ConcreteModel()
    nt = 11  # Eleven timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    with pytest.raises(
        ConfigurationError, match="User specified antiwindup method for unbounded MV."
    ):
        unit = m.fs.unit = PIDController(
            process_var=m.fs.process_var,
            manipulated_var=m.fs.manipulated_var,
            controller_type=ControllerType.PID,
            mv_bound_type=ControllerMVBoundType.NONE,
            antiwindup_type=ControllerAntiwindupType.BACK_CALCULATION,
        )


@pytest.mark.unit
def test_antiwindup_no_integration():
    m = ConcreteModel()
    nt = 7  # Five timepoints
    m.fs = FlowsheetBlock(dynamic=True, time_set=range(nt), time_units=pyunits.s)

    m.fs.process_var = Var(m.fs.time, units=pyunits.m)
    m.fs.manipulated_var = Var(m.fs.time, units=pyunits.K)

    with pytest.raises(
        ConfigurationError,
        match="User specified antiwindup method for controller without integral action.",
    ):
        unit = m.fs.unit = PIDController(
            process_var=m.fs.process_var,
            manipulated_var=m.fs.manipulated_var,
            controller_type=ControllerType.PD,
            mv_bound_type=ControllerMVBoundType.SMOOTH_BOUND,
            antiwindup_type=ControllerAntiwindupType.BACK_CALCULATION,
        )


# if __name__ == "__main__":
#     test_config()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for IAPWS using a heater block.  This is an easy way to test the
different calculation method options.

Author: John Eslick
"""
import pytest

from pyomo.environ import ConcreteModel, value

from idaes.core import FlowsheetBlock
from idaes.models.unit_models import Heater
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core import MaterialBalanceType
from idaes.core.solvers import get_solver
from idaes.models.properties.general_helmholtz import (
    HelmholtzParameterBlock,
    PhaseType,
    StateVars,
)
from idaes.models.properties.general_helmholtz import helmholtz_available


# -----------------------------------------------------------------------------
# set up solver
solver = get_solver("ipopt_v2")


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_heater_ph_mixed_byphase():
    """Test mixed phase form with P-H state vars and phase mass balances"""
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = HelmholtzParameterBlock(pure_component="h2o")
    m.fs.heater = Heater(property_package=m.fs.properties)
    m.fs.heater.inlet.enth_mol[0].set_value(4000)
    m.fs.heater.inlet.flow_mol[0].fix(100)
    m.fs.heater.inlet.pressure[0].set_value(101325)
    m.fs.heater.heat_duty[0].fix(100 * 20000)

    # Test that property block properly holds state during initialization
    flags = m.fs.heater.control_volume.initialize()
    assert degrees_of_freedom(m) == 0

    # Make sure original states are restored
    m.fs.heater.control_volume.release_state(flags)
    assert not m.fs.heater.inlet.enth_mol[0].fixed
    assert m.fs.heater.inlet.flow_mol[0].fixed
    assert not m.fs.heater.inlet.pressure[0].fixed

    m.fs.heater.inlet.enth_mol.fix()
    m.fs.heater.inlet.pressure.fix()

    m.fs.heater.initialize()
    assert degrees_of_freedom(m) == 0
    solver.solve(m)
    prop_in = m.fs.heater.control_volume.properties_in[0]
    prop_out = m.fs.heater.control_volume.properties_out[0]
    assert value(prop_in.temperature) == pytest.approx(326.166707, rel=1e-5)
    assert value(prop_out.temperature) == pytest.approx(373.12429, rel=1e-5)
    assert value(prop_in.phase_frac["Liq"]) == pytest.approx(1, rel=1e-6)
    assert value(prop_out.phase_frac["Liq"]) == pytest.approx(0.5953219, rel=1e-5)
    assert value(prop_in.phase_frac["Vap"]) == pytest.approx(0, abs=1e-6)
    assert value(prop_out.phase_frac["Vap"]) == pytest.approx(0.4046781, rel=1e-4)
    assert value(prop_out.mole_frac_comp["h2o"]) == pytest.approx(1.0, rel=1e-4)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_heater_phmixed_mixed_total():
    """Test mixed phase form with P-H state vars and total mass balances"""
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = HelmholtzParameterBlock(pure_component="h2o")
    m.fs.heater = Heater(
        property_package=m.fs.properties,
        material_balance_type=MaterialBalanceType.componentTotal,
    )
    m.fs.heater.inlet.enth_mol.fix(4000)
    m.fs.heater.inlet.flow_mol.fix(100)
    m.fs.heater.inlet.pressure.fix(101325)
    m.fs.heater.heat_duty[0].fix(100 * 20000)
    m.fs.heater.initialize()
    assert degrees_of_freedom(m) == 0
    solver.solve(m)
    prop_in = m.fs.heater.control_volume.properties_in[0]
    prop_out = m.fs.heater.control_volume.properties_out[0]
    assert value(prop_in.temperature) == pytest.approx(326.166707, rel=1e-5)
    assert value(prop_out.temperature) == pytest.approx(373.12429, rel=1e-5)
    assert value(prop_in.phase_frac["Liq"]) == pytest.approx(1, rel=1e-6)
    assert value(prop_out.phase_frac["Liq"]) == pytest.approx(0.5953219, rel=1e-5)
    assert value(prop_in.phase_frac["Vap"]) == pytest.approx(0, abs=1e-5)
    assert value(prop_out.phase_frac["Vap"]) == pytest.approx(0.4046781, rel=1e-4)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_heater_ph_lg_total():
    """Test liquid/vapor form with P-H state vars and total mass balances"""
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = HelmholtzParameterBlock(
        pure_component="h2o", phase_presentation=PhaseType.LG
    )
    m.fs.heater = Heater(
        property_package=m.fs.properties,
        material_balance_type=MaterialBalanceType.componentTotal,
    )
    m.fs.heater.inlet.enth_mol.fix(4000)
    m.fs.heater.inlet.flow_mol.fix(100)
    m.fs.heater.inlet.pressure.fix(101325)
    m.fs.heater.heat_duty[0].fix(100 * 20000)
    m.fs.heater.initialize()
    assert degrees_of_freedom(m) == 0
    solver.solve(m)
    prop_in = m.fs.heater.control_volume.properties_in[0]
    prop_out = m.fs.heater.control_volume.properties_out[0]
    assert value(prop_in.temperature) == pytest.approx(326.166707, rel=1e-5)
    assert value(prop_out.temperature) == pytest.approx(373.12429, rel=1e-5)
    assert value(prop_in.phase_frac["Liq"]) == pytest.approx(1, rel=1e-5)
    assert value(prop_out.phase_frac["Liq"]) == pytest.approx(0.5953219, rel=1e-5)
    assert value(prop_in.phase_frac["Vap"]) == pytest.approx(0, abs=1e-5)
    assert value(prop_out.phase_frac["Vap"]) == pytest.approx(0.4046781, rel=1e-4)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_heater_ph_lg_phase():
    """Test liquid/vapor form with P-H state vars and phase mass balances

    This is known not to work so this is a place holder to demonstrate the extra
    equation and reserve a place for a test once the problem is fixed. The
    issue in case I forget is that the phase based mass balance writes a mass
    balnace for each phase, but the P-H formulation always calculates a phase
    fraction making one equation redundant.
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = HelmholtzParameterBlock(
        pure_component="h2o", phase_presentation=PhaseType.LG
    )
    m.fs.heater = Heater(
        property_package=m.fs.properties,
        material_balance_type=MaterialBalanceType.componentPhase,
    )
    m.fs.heater.inlet.enth_mol.fix(4000)
    m.fs.heater.inlet.flow_mol.fix(100)
    m.fs.heater.inlet.pressure.fix(101325)
    m.fs.heater.heat_duty[0].fix(100 * 20000)
    assert degrees_of_freedom(m) == -1


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_heater_ph_l_phase_two():
    """Test liquid phase only form with P-H state vars and phase mass balances
    where the result is a two phase mixture.  For the P-H vars if there are
    two phases, using the liquid only properties will case the phase fraction
    to report all liquid even though really it is not.  Only use this option
    if you are sure you do not have a vapor phase.
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = HelmholtzParameterBlock(
        pure_component="h2o", phase_presentation=PhaseType.L
    )
    m.fs.heater = Heater(
        property_package=m.fs.properties,
        material_balance_type=MaterialBalanceType.componentPhase,
    )
    m.fs.heater.inlet.enth_mol.fix(3000)
    m.fs.heater.inlet.flow_mol.fix(100)
    m.fs.heater.inlet.pressure.fix(101325)
    m.fs.heater.heat_duty[0].fix(100 * 1000)
    m.fs.heater.initialize()
    assert degrees_of_freedom(m) == 0
    solver.solve(m)
    prop_in = m.fs.heater.control_volume.properties_in[0]
    prop_out = m.fs.heater.control_volume.properties_out[0]
    assert abs(value(prop_in.temperature) - 312.88896252921666) <= 1e-4
    assert abs(value(prop_out.temperature) - 326.166707507874) <= 1e-4
    assert abs(value(prop_in.phase_frac["Liq"]) - 1) <= 1e-6
    assert abs(value(prop_out.phase_frac["Liq"]) - 1) <= 1e-6


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_heater_ph_l_phase():
    """Test liquid phase only form with P-H state vars and phase mass balances"""
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = HelmholtzParameterBlock(
        pure_component="h2o",
        phase_presentation=PhaseType.L,
    )
    m.fs.heater = Heater(
        property_package=m.fs.properties,
        material_balance_type=MaterialBalanceType.componentPhase,
    )
    m.fs.heater.inlet.enth_mol.fix(3000)
    m.fs.heater.inlet.flow_mol.fix(100)
    m.fs.heater.inlet.pressure.fix(101325)
    m.fs.heater.heat_duty[0].fix(100 * 1000)
    m.fs.heater.initialize()
    assert degrees_of_freedom(m) == 0
    solver.solve(m)
    prop_in = m.fs.heater.control_volume.properties_in[0]
    prop_out = m.fs.heater.control_volume.properties_out[0]
    assert abs(value(prop_in.temperature) - 312.88896252921666) <= 1e-4
    assert abs(value(prop_out.temperature) - 326.166707507874) <= 1e-4
    assert abs(value(prop_in.phase_frac["Liq"]) - 1) <= 1e-6
    assert abs(value(prop_out.phase_frac["Liq"]) - 1) <= 1e-6


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_heater_ph_g_phase():
    """Test vapor phase only form with P-H state vars and phase mass balances"""
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = HelmholtzParameterBlock(
        pure_component="h2o", phase_presentation=PhaseType.G
    )
    m.fs.heater = Heater(
        property_package=m.fs.properties,
        material_balance_type=MaterialBalanceType.componentPhase,
    )
    m.fs.heater.inlet.enth_mol.fix(50000)
    m.fs.heater.inlet.flow_mol.fix(100)
    m.fs.heater.inlet.pressure.fix(101325)
    m.fs.heater.heat_duty[0].fix(100 * 10000)
    m.fs.heater.initialize()
    assert degrees_of_freedom(m) == 0
    solver.solve(m)
    prop_in = m.fs.heater.control_volume.properties_in[0]
    prop_out = m.fs.heater.control_volume.properties_out[0]
    assert abs(value(prop_in.temperature) - 422.60419933276177) <= 1e-2
    assert abs(value(prop_out.temperature) - 698.1604861702295) <= 1e-2
    assert abs(value(prop_in.phase_frac["Vap"]) - 1) <= 1e-6
    assert abs(value(prop_out.phase_frac["Vap"]) - 1) <= 1e-6


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_heater_tpx_g_phase():
    """Test vapor phase only form with T-P-x state vars and phase mass balances"""
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = HelmholtzParameterBlock(
        pure_component="h2o",
        phase_presentation=PhaseType.G,
        state_vars=StateVars.TPX,
    )
    m.fs.heater = Heater(
        property_package=m.fs.properties,
        material_balance_type=MaterialBalanceType.componentPhase,
    )
    m.fs.heater.inlet.flow_mol.fix(100)
    m.fs.heater.inlet.temperature.fix(422.60419933276177)
    m.fs.heater.inlet.pressure.fix(101325)
    m.fs.heater.heat_duty[0].fix(100 * 10000)
    prop_in = m.fs.heater.control_volume.properties_in[0]
    prop_out = m.fs.heater.control_volume.properties_out[0]
    prop_out.temperature = 550
    prop_out.pressure = 101325
    prop_out.flow_mol = 100
    m.fs.heater.initialize(outlvl=5)
    assert degrees_of_freedom(m) == 0
    solver.solve(m)
    assert abs(value(prop_out.temperature) - 698.1604861702295) <= 1e-3
    assert abs(value(prop_in.phase_frac["Vap"]) - 1) <= 1e-6
    assert abs(value(prop_out.phase_frac["Vap"]) - 1) <= 1e-6


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_heater_tpx_lg_total():
    """Test liquid/vapor form with T-P-x state vars and total mass balances. In
    this case you end up with two-phases at the end.
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = HelmholtzParameterBlock(
        pure_component="h2o",
        phase_presentation=PhaseType.LG,
        state_vars=StateVars.TPX,
    )
    m.fs.heater = Heater(
        property_package=m.fs.properties,
        material_balance_type=MaterialBalanceType.componentTotal,
    )
    m.fs.heater.inlet.temperature.fix(326.1667075078748)
    m.fs.heater.inlet.vapor_frac.fix(0.0)
    m.fs.heater.inlet.flow_mol.fix(100)
    m.fs.heater.inlet.pressure.fix(101325)
    m.fs.heater.heat_duty[0].fix(100 * 20000)
    prop_in = m.fs.heater.control_volume.properties_in[0]
    prop_out = m.fs.heater.control_volume.properties_out[0]
    prop_out.temperature = 400
    prop_out.vapor_frac = 0.9
    m.fs.heater.initialize(outlvl=5)
    assert degrees_of_freedom(m) == 0
    solver.solve(m, tee=True)
    assert value(prop_in.temperature) == pytest.approx(326.166707, rel=1e-4)
    assert value(prop_out.temperature) == pytest.approx(373.12429, rel=1e-4)
    assert value(prop_in.phase_frac["Liq"]) == pytest.approx(1, rel=1e-5)
    assert value(prop_out.phase_frac["Liq"]) == pytest.approx(0.5953219, rel=1e-5)
    assert value(prop_in.phase_frac["Vap"]) == pytest.approx(0, abs=1e-5)
    assert value(prop_out.phase_frac["Vap"]) == pytest.approx(0.4046781, rel=1e-4)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_heater_tpx_lg_total_2():
    """Test liquid/vapor form with T-P-x state vars and total mass balances. In
    this case you end up with all vapor at the end.
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = HelmholtzParameterBlock(
        pure_component="h2o",
        phase_presentation=PhaseType.LG,
        state_vars=StateVars.TPX,
    )
    m.fs.heater = Heater(
        property_package=m.fs.properties,
        material_balance_type=MaterialBalanceType.componentTotal,
    )
    m.fs.heater.inlet.temperature.fix(326.1667075078748)
    m.fs.heater.inlet.vapor_frac.fix(0.0)
    m.fs.heater.inlet.flow_mol.fix(100)
    m.fs.heater.inlet.pressure.fix(101325)
    m.fs.heater.heat_duty[0].fix(100 * 50000)
    prop_in = m.fs.heater.control_volume.properties_in[0]
    prop_out = m.fs.heater.control_volume.properties_out[0]
    prop_out.temperature = 600
    prop_out.vapor_frac = 0.9
    m.fs.heater.initialize()
    assert degrees_of_freedom(m) == 0
    solver.solve(m, tee=True)
    assert abs(value(prop_out.temperature) - 534.6889772922356) <= 1e-3
    assert abs(value(prop_in.phase_frac["Liq"]) - 1) <= 1e-2
    assert abs(value(prop_out.phase_frac["Liq"]) - 0) <= 1e-2
    assert abs(value(prop_in.phase_frac["Vap"]) - 0) <= 1e-2
    assert abs(value(prop_out.phase_frac["Vap"]) - 1) <= 1e-2


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_heater_tpx_lg_phase():
    """Test liquid/vapor form with T-P-x state vars and phase mass balances.

    This again has the problem where the property package calculates a vapor
    fraction, but you have a mass balance for each phase.
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = HelmholtzParameterBlock(
        pure_component="h2o",
        phase_presentation=PhaseType.LG,
        state_vars=StateVars.TPX,
    )
    m.fs.heater = Heater(
        property_package=m.fs.properties,
        material_balance_type=MaterialBalanceType.componentPhase,
    )
    m.fs.heater.inlet.temperature.fix(326.1667075078748)
    m.fs.heater.inlet.vapor_frac.fix(0.0)
    m.fs.heater.inlet.flow_mol.fix(100)
    m.fs.heater.inlet.pressure.fix(101325)
    m.fs.heater.heat_duty[0].fix(100 * 50000)

    assert degrees_of_freedom(m) == -1



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
1-D Electric Trim Heater Model With Wall Temperatures

Discretization based on tube rows
"""
# Import Pyomo libraries
from pyomo.environ import (
    assert_optimal_termination,
    Block,
    units as pyunits,
    Var,
    value,
)
from pyomo.common.config import ConfigValue, In
from pyomo.util.calc_var_value import calculate_variable_from_constraint

# Import IDAES cores
from idaes.core import (
    ControlVolume1DBlock,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    FlowDirection,
    UnitModelBlockData,
    useDefault,
)
import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import InitializationError
from idaes.core.util.misc import add_object_reference
import idaes.logger as idaeslog
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.models_extra.power_generation.unit_models.heat_exchanger_common import (
    make_geometry_common,
    make_performance_common,
    scale_common,
)
from idaes.core.initialization import SingleControlVolumeUnitInitializer

__author__ = "Jinliang Ma, Douglas Allan"


class Heater1DInitializer(SingleControlVolumeUnitInitializer):
    """
    Initializer for Heater 1D units.

    A simple initialization method that first initializes the control volume
    without heat transfer, then adds heat transfer in and solves it again,
    then finally solves the entire model.
    """

    def initialize_main_model(
        self,
        model: Block,
        copy_inlet_state: bool = False,
    ):
        """
        Initialization routine for the main Heater 1D model
        (as opposed to submodels like costing, which presently do not exist).


        Args:
            model: Pyomo Block to be initialized.
            copy_inlet_state: bool (default=False). Whether to copy inlet state to other states or not
                (0-D control volumes only). Copying will generally be faster, but inlet states may not contain
                all properties required elsewhere.
            duty: initial guess for heat duty to assist with initialization. Can be a Pyomo expression with units.

        Returns:
            Pyomo solver results object.

        """
        # Set solver options
        init_log = idaeslog.getInitLogger(
            model.name, self.get_output_level(), tag="unit"
        )
        solve_log = idaeslog.getSolveLogger(
            model.name, self.get_output_level(), tag="unit"
        )

        # Create solver
        solver_obj = get_solver(self.config.solver, self.config.solver_options)

        # ---------------------------------------------------------------------
        # Initialize shell block

        self.initialize_control_volume(model.control_volume, copy_inlet_state)

        init_log.info_high("Initialization Step 1 Complete.")

        calc_var = calculate_variable_from_constraint

        calc_var(model.length_flow_shell, model.length_flow_shell_eqn)
        calc_var(model.area_flow_shell, model.area_flow_shell_eqn)
        calc_var(model.area_flow_shell_min, model.area_flow_shell_min_eqn)

        for t in model.flowsheet().config.time:
            for x in model.control_volume.length_domain:
                model.control_volume.heat[t, x].fix(
                    value(model.electric_heat_duty[t] / model.length_flow_shell)
                )

        if model.config.has_pressure_change:
            model.control_volume.pressure.fix()

        model.control_volume.length.fix()
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver_obj.solve(model.control_volume, tee=slc.tee)

        try:
            assert_optimal_termination(res)
        except AssertionError:
            raise InitializationError("Initialization solve failed.")

        init_log.info_high("Initialization Step 2 Complete.")
        model.control_volume.length.unfix()
        model.control_volume.heat.unfix()

        for t in model.flowsheet().config.time:
            for x in model.control_volume.length_domain:
                model.temp_wall_center[t, x].fix(
                    value(model.control_volume.properties[t, x].temperature) + 10
                )
                calc_var(model.heat_holdup[t, x], model.heat_holdup_eqn[t, x])
                model.temp_wall_center[t, x].unfix()

        if model.config.has_pressure_change:
            model.control_volume.pressure.unfix()
            model.control_volume.pressure[:, 0].fix()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver_obj.solve(model, tee=slc.tee)

        try:
            assert_optimal_termination(res)
        except AssertionError:
            raise InitializationError("Initialization solve failed.")

        init_log.info_high("Initialization Step 3 Complete.")

        return res


@declare_process_block_class("Heater1D")
class Heater1DData(UnitModelBlockData):
    """Standard Trim Heater Model Class Class."""

    default_initializer = Heater1DInitializer

    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "has_fluid_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag",
            doc="""Indicates whether holdup terms for the fluid should be constructed or not.
            **default** - False.
            **Valid values:** {
            **False** - do not construct holdup terms}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.componentTotal,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.componentTotal.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.enthalpyTotal,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a ParameterBlock object""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigValue(
            default=None,
            description="Arguments for constructing shell property package",
            doc="""A dict of arguments to be passed to the PropertyBlockData
and used when constructing these
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a dict (see property package for documentation)""",
        ),
    )
    CONFIG.declare(
        "transformation_method",
        ConfigValue(
            default=useDefault,
            description="Discretization method to use for DAE transformation",
            doc="""Discretization method to use for DAE transformation. See Pyomo
documentation for supported transformations.""",
        ),
    )
    CONFIG.declare(
        "transformation_scheme",
        ConfigValue(
            default=useDefault,
            description="Discretization scheme to use for DAE transformation",
            doc="""Discretization scheme to use when transforming domain. See Pyomo
documentation for supported schemes.""",
        ),
    )

    # Common config args for both sides
    CONFIG.declare(
        "finite_elements",
        ConfigValue(
            default=5,
            domain=int,
            description="Number of finite elements length domain",
            doc="""Number of finite elements to use when discretizing length
domain (default=5). Should set to the number of tube rows""",
        ),
    )
    CONFIG.declare(
        "collocation_points",
        ConfigValue(
            default=3,
            domain=int,
            description="Number of collocation points per finite element",
            doc="""If using collocation, number of collocation points to use
            per finite element when discretizing length domain (default=3)""",
        ),
    )
    CONFIG.declare(
        "tube_arrangement",
        ConfigValue(
            default="in-line",
            domain=In(["in-line", "staggered"]),
            description="tube configuration",
            doc="tube arrangement could be in-line or staggered",
        ),
    )

    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super().build()

        # Set flow directions for the control volume blocks and specify
        # dicretization if not specified.
        set_direction_shell = FlowDirection.forward
        if self.config.transformation_method is useDefault:
            self.config.transformation_method = "dae.finite_difference"
        if self.config.transformation_scheme is useDefault:
            self.config.transformation_scheme = "BACKWARD"

        if self.config.property_package_args is None:
            self.config.property_package_args = {}

        # Control volume 1D for shell, set to steady-state for fluid
        self.control_volume = ControlVolume1DBlock(
            dynamic=self.config.dynamic and self.config.has_fluid_holdup,
            has_holdup=self.config.has_fluid_holdup,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
            transformation_method=self.config.transformation_method,
            transformation_scheme=self.config.transformation_scheme,
            finite_elements=self.config.finite_elements,
            collocation_points=self.config.collocation_points,
        )

        self.control_volume.add_geometry(flow_direction=set_direction_shell)

        self.control_volume.add_state_blocks(
            information_flow=set_direction_shell,
            has_phase_equilibrium=False,
        )

        # Populate shell
        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_phase_equilibrium=False,
        )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type, has_heat_transfer=True
        )

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change,
        )

        self.control_volume.apply_transformation()

        # Populate tube

        # Add Ports for shell side
        self.add_inlet_port(name="inlet", block=self.control_volume)
        self.add_outlet_port(name="outlet", block=self.control_volume)

        self._make_geometry()

        self._make_performance()

    def _make_geometry(self):
        """
        Constraints for unit model.

        Args:
            None

        Returns:
            None
        """
        units = self.config.property_package.get_metadata().derived_units
        # Add reference to control volume geometry
        add_object_reference(self, "area_flow_shell", self.control_volume.area)
        add_object_reference(self, "length_flow_shell", self.control_volume.length)
        make_geometry_common(self, shell_units=units)

        @self.Expression(
            doc="Common performance equations expect this expression to be here"
        )
        def length_flow_tube(b):
            return b.number_passes * b.length_tube_seg

    def _make_performance(self):
        """
        Constraints for unit model.

        Args:
            None

        Returns:
            None
        """
        self.electric_heat_duty = Var(
            self.flowsheet().config.time,
            initialize=1e6,
            units=pyunits.W,
            doc="Heat duty provided to heater through resistive heating",
        )
        units = self.config.property_package.get_metadata().derived_units
        make_performance_common(
            self,
            shell=self.control_volume,
            shell_units=units,
            shell_has_pressure_change=self.config.has_pressure_change,
            make_reynolds=True,
            make_nusselt=True,
        )

        def heat_accumulation_term(b, t, x):
            return b.heat_accumulation[t, x] if b.config.dynamic else 0

        # Nusselt number, currently assume Re>300
        @self.Constraint(
            self.flowsheet().config.time,
            self.control_volume.length_domain,
            doc="Nusselts number equation",
        )
        def N_Nu_shell_eqn(b, t, x):
            return (
                b.N_Nu_shell[t, x]
                == b.f_arrangement
                * 0.33
                * b.N_Re_shell[t, x] ** 0.6
                * b.control_volume.properties[t, x].prandtl_number_phase["Vap"]
                ** 0.333333
            )

        # Energy balance with tube wall
        # ------------------------------------
        # Heat to wall per length
        @self.Constraint(
            self.flowsheet().config.time,
            self.control_volume.length_domain,
            doc="heat per length",
        )
        def heat_shell_eqn(b, t, x):
            return b.control_volume.heat[t, x] * b.length_flow_shell == (
                b.total_heat_transfer_coeff_shell[t, x]
                * b.total_heat_transfer_area
                * (
                    b.temp_wall_shell[t, x]
                    - b.control_volume.properties[t, x].temperature
                )
            )

        # Shell side wall temperature
        @self.Constraint(
            self.flowsheet().config.time,
            self.control_volume.length_domain,
            doc="shell side wall temperature",
        )
        def temp_wall_shell_eqn(b, t, x):
            return (
                b.total_heat_transfer_coeff_shell[t, x]
                * (
                    b.control_volume.properties[t, x].temperature
                    - b.temp_wall_shell[t, x]
                )
                # Divide thickness by 2 in order to represent center of hollow tube instead of
                # interior edge of hollow tube
                * (b.thickness_tube / (2 * b.therm_cond_wall) + b.rfouling_shell)
                == b.temp_wall_shell[t, x] - b.temp_wall_center[t, x]
            )

        @self.Constraint(
            self.flowsheet().config.time,
            self.control_volume.length_domain,
            doc="wall temperature",
        )
        def temp_wall_center_eqn(b, t, x):
            return heat_accumulation_term(b, t, x) == (
                -b.control_volume.heat[t, x]
                + b.electric_heat_duty[t] / b.length_flow_shell
            )

    def calculate_scaling_factors(self):
        def gsf(obj):
            return iscale.get_scaling_factor(obj, default=1, warning=True)

        def ssf(obj, sf):
            iscale.set_scaling_factor(obj, sf, overwrite=False)

        def cst(con, sf):
            iscale.constraint_scaling_transform(con, sf, overwrite=False)

        scale_common(
            self,
            self.control_volume,
            self.config.has_pressure_change,
            make_reynolds=True,
            make_nusselt=True,
        )

        sf_d_tube = iscale.get_scaling_factor(
            self.do_tube, default=1 / value(self.do_tube)
        )

        for t in self.flowsheet().time:
            for z in self.control_volume.length_domain:
                sf_hconv_conv = gsf(self.conv_heat_transfer_coeff_shell[t, z])
                cst(
                    self.conv_heat_transfer_coeff_shell_eqn[t, z],
                    sf_hconv_conv * sf_d_tube,
                )

                sf_T = gsf(self.control_volume.properties[t, z].temperature)
                ssf(self.temp_wall_shell[t, z], sf_T)
                ssf(self.temp_wall_center[t, z], sf_T)

                s_Q = gsf(self.control_volume.heat[t, z])
                ssf(self.electric_heat_duty[t], s_Q / value(self.length_flow_shell))
                cst(self.heat_shell_eqn[t, z], s_Q * value(self.length_flow_shell))
                ssf(self.temp_wall_center[t, z], sf_T)
                cst(self.temp_wall_shell_eqn[t, z], sf_T)
                cst(self.temp_wall_center_eqn[t, z], s_Q)

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        var_dict["Electric Heat Duty"] = self.electric_heat_duty[time_point]

        expr_dict = {}
        expr_dict["HX Area"] = self.total_heat_transfer_area

        return {"vars": var_dict, "exprs": expr_dict}

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Inlet": self.inlet,
                "Outlet": self.outlet,
            },
            time_point=time_point,
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Surrogate Plotting Methods
"""

import pytest
import os
import pandas as pd
import matplotlib.pyplot as plt

from pyomo.common.fileutils import this_file_dir
from pyomo.common.tempfiles import TempfileManager

from idaes.core.surrogate.alamopy import AlamoSurrogate
from idaes.core.surrogate.sampling.data_utils import split_training_validation
from idaes.core.surrogate.plotting.sm_plotter import (
    surrogate_scatter2D,
    surrogate_scatter3D,
    surrogate_parity,
    surrogate_residual,
)


@pytest.fixture
def alamo_surrogate():
    # import surrogates from external JSON

    alamo_surrogate = AlamoSurrogate.load_from_file(
        os.path.join(this_file_dir(), "alamo_surrogate.json")
    )

    return alamo_surrogate


@pytest.fixture
def data_validation():
    # import validation data and create dataframe here as well
    # random data subset will change every time, but should be fine

    csv_data = pd.read_csv(os.path.join(this_file_dir(), "reformer-data.csv"))
    data = csv_data.sample(n=100)  # randomly sample points for validation

    input_data = data.iloc[:, :2]
    input_labels = input_data.columns

    n_data = data[input_labels[0]].size  # size = 100
    data_training, data_validation = split_training_validation(data, 0.8, seed=n_data)

    return data_validation


# --------------------------------------------------------------------------


@pytest.mark.unit
def test_scatter2D_alamo(alamo_surrogate, data_validation):

    with TempfileManager.new_context() as tf:
        # note - a failure 'The process cannot access the file because it is
        # being used by another process' could occur if an internal error
        # arises before the results file is closed inside the surrogate method

        # create and step into new temporary directory
        dname = tf.mkdtemp()
        filename = os.path.join(dname, "results.pdf")
        figs = surrogate_scatter2D(
            alamo_surrogate, data_validation, filename=filename, show=False
        )
        for f in figs:
            plt.close(f)

        assert os.path.exists(filename)  # PDF results file


@pytest.mark.unit
def test_scatter3D_alamo(alamo_surrogate, data_validation):

    with TempfileManager.new_context() as tf:
        # note - a failure 'The process cannot access the file because it is
        # being used by another process' could occur if an internal error
        # arises before the results file is closed inside the surrogate method

        # create and step into new temporary directory
        dname = tf.mkdtemp()
        filename = os.path.join(dname, "results.pdf")
        figs = surrogate_scatter3D(
            alamo_surrogate, data_validation, filename=filename, show=False
        )
        for f in figs:
            plt.close(f)

        assert os.path.exists(filename)  # PDF results file


@pytest.mark.unit
def test_parity_alamo(alamo_surrogate, data_validation):

    with TempfileManager.new_context() as tf:
        # note - a failure 'The process cannot access the file because it is
        # being used by another process' could occur if an internal error
        # arises before the results file is closed inside the surrogate method

        # create and step into new temporary directory
        dname = tf.mkdtemp()
        filename = os.path.join(dname, "results.pdf")
        figs = surrogate_parity(
            alamo_surrogate, data_validation, filename=filename, show=False
        )
        for f in figs:
            plt.close(f)

        assert os.path.exists(filename)  # PDF results file


@pytest.mark.unit
def test_residual_alamo(alamo_surrogate, data_validation):

    with TempfileManager.new_context() as tf:
        # note - a failure 'The process cannot access the file because it is
        # being used by another process' could occur if an internal error
        # arises before the results file is closed inside the surrogate method

        # create and step into new temporary directory
        dname = tf.mkdtemp()
        filename = os.path.join(dname, "results.pdf")
        figs = surrogate_residual(
            alamo_surrogate, data_validation, filename=filename, show=False
        )
        for f in figs:
            plt.close(f)

        assert os.path.exists(filename)  # PDF results file


@pytest.mark.unit
def test_scatter2D_noPDF_alamo(alamo_surrogate, data_validation):

    with TempfileManager.new_context() as tf:
        # note - a failure 'The process cannot access the file because it is
        # being used by another process' could occur if an internal error
        # arises before the results file is closed inside the surrogate method

        # create and step into new temporary directory
        dname = tf.mkdtemp()
        figs = surrogate_scatter2D(alamo_surrogate, data_validation, show=False)
        for f in figs:
            plt.close(f)

        for file in list(os.walk(dname)):  # check entire temp directory
            assert file[-4:] != ".pdf"  # no PDF files should be created


@pytest.mark.unit
def test_scatter3D_noPDF_alamo(alamo_surrogate, data_validation):

    with TempfileManager.new_context() as tf:
        # note - a failure 'The process cannot access the file because it is
        # being used by another process' could occur if an internal error
        # arises before the results file is closed inside the surrogate method

        # create and step into new temporary directory
        dname = tf.mkdtemp()
        figs = surrogate_scatter3D(alamo_surrogate, data_validation, show=False)
        for f in figs:
            plt.close(f)

        for file in list(os.walk(dname)):  # check entire temp directory
            assert file[-4:] != ".pdf"  # no PDF files should be created


@pytest.mark.unit
def test_parity_noPDF_alamo(alamo_surrogate, data_validation):

    with TempfileManager.new_context() as tf:
        # note - a failure 'The process cannot access the file because it is
        # being used by another process' could occur if an internal error
        # arises before the results file is closed inside the surrogate method

        # create and step into new temporary directory
        dname = tf.mkdtemp()
        figs = surrogate_parity(alamo_surrogate, data_validation, show=False)
        for f in figs:
            plt.close(f)

        for file in list(os.walk(dname)):  # check entire temp directory
            assert file[-4:] != ".pdf"  # no PDF files should be created


@pytest.mark.unit
def test_residual_noPDF_alamo(alamo_surrogate, data_validation):

    with TempfileManager.new_context() as tf:
        # note - a failure 'The process cannot access the file because it is
        # being used by another process' could occur if an internal error
        # arises before the results file is closed inside the surrogate method

        # create and step into new temporary directory
        dname = tf.mkdtemp()
        figs = surrogate_residual(alamo_surrogate, data_validation, show=False)
        for f in figs:
            plt.close(f)

        for file in list(os.walk(dname)):  # check entire temp directory
            assert file[-4:] != ".pdf"  # no PDF files should be created



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for valve model
"""
import math
import pytest

from pyomo.environ import check_optimal_termination, ConcreteModel, units, value
from idaes.core import FlowsheetBlock
from idaes.models.unit_models import Valve, ValveFunctionType
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.util.testing import (
    initialization_tester,
)
from idaes.core.solvers import get_solver
from pyomo.util.check_units import assert_units_consistent, assert_units_equivalent

from idaes.models.properties import iapws95
import idaes.core.util.scaling as iscale
from idaes.models.properties.general_helmholtz import helmholtz_available
from idaes.core.util import DiagnosticsToolbox

# -----------------------------------------------------------------------------
# Get default solver for testing
# TODO: Will need to switch this to not use user-scaling once scaling tools are ready
solver = get_solver(
    solver="ipopt_v2", solver_options={"nlp_scaling_method": "user-scaling"}
)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class GenericValve(object):
    @pytest.fixture(scope="class")
    def valve_model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = iapws95.Iapws95ParameterBlock()
        m.fs.valve = Valve(
            valve_function_callback=self.type, property_package=m.fs.properties
        )
        fin = 1000  # mol/s
        pin = 200000  # Pa
        pout = 100000  # Pa
        tin = 300  # K
        hin = iapws95.htpx(T=tin * units.K, P=pin * units.Pa)  # J/mol
        # Calculate the flow coefficient to give 1000 mol/s flow with given P
        if self.type == ValveFunctionType.linear:
            cv = 1000 / math.sqrt(pin - pout) / 0.5
        elif self.type == ValveFunctionType.quick_opening:
            cv = 1000 / math.sqrt(pin - pout) / math.sqrt(0.5)
        elif self.type == ValveFunctionType.equal_percentage:
            cv = 1000 / math.sqrt(pin - pout) / 100 ** (0.5 - 1)
        # set inlet
        m.fs.valve.inlet.enth_mol[0].fix(hin)
        m.fs.valve.inlet.flow_mol[0].fix(fin)
        m.fs.valve.inlet.pressure[0].fix(pin)
        m.fs.valve.outlet.pressure[0].set_value(pout)
        m.fs.valve.Cv.fix(cv)
        m.fs.valve.valve_opening.fix(0.5)
        iscale.calculate_scaling_factors(m)
        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, valve_model):
        assert hasattr(valve_model.fs.valve, "inlet")
        assert len(valve_model.fs.valve.inlet.vars) == 3
        assert hasattr(valve_model.fs.valve.inlet, "flow_mol")
        assert hasattr(valve_model.fs.valve.inlet, "enth_mol")
        assert hasattr(valve_model.fs.valve.inlet, "pressure")

        assert hasattr(valve_model.fs.valve, "outlet")
        assert len(valve_model.fs.valve.outlet.vars) == 3
        assert hasattr(valve_model.fs.valve.outlet, "flow_mol")
        assert hasattr(valve_model.fs.valve.outlet, "enth_mol")
        assert hasattr(valve_model.fs.valve.outlet, "pressure")

        assert hasattr(valve_model.fs.valve, "pressure_flow_equation")
        assert hasattr(valve_model.fs.valve, "valve_opening")
        assert hasattr(valve_model.fs.valve, "valve_function")
        assert hasattr(valve_model.fs.valve, "flow_var")

        if self.type == ValveFunctionType.equal_percentage:
            # alpha valve function parameter is one more than others
            assert number_variables(valve_model) == 12
        else:
            assert number_variables(valve_model) == 11

        assert number_total_constraints(valve_model) == 6
        assert number_unused_variables(valve_model) == 0

    @pytest.mark.component
    def test_structural_issues(self, valve_model):
        dt = DiagnosticsToolbox(valve_model)
        dt.assert_no_structural_warnings()

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, valve_model):
        initialization_tester(valve_model, unit=valve_model.fs.valve)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, valve_model):
        results = solver.solve(valve_model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, valve_model):
        # calculated Cv to yield this solution
        assert pytest.approx(1000, rel=1e-3) == value(
            valve_model.fs.valve.outlet.flow_mol[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, valve_model):
        dt = DiagnosticsToolbox(valve_model)
        dt.assert_no_numerical_warnings()


class TestLinearValve(GenericValve):
    type = ValveFunctionType.linear


class TestEPValve(GenericValve):
    type = ValveFunctionType.equal_percentage


class TestQuickValve(GenericValve):
    type = ValveFunctionType.quick_opening
